from typing import List

import numpy as np


def get_class_count(truth_index: List[List[int]],
                    truth: np.ndarray,
                    class_weights_threshold: List[float]) -> List[List[int]]:
    class_count = [[] for _ in range(len(truth_index))]
    for n in range(len(truth_index)):
        class_count[n] = [0] * len(truth_index[n])
        for idx, cl in enumerate(truth_index[n]):
            truth_sum = int(np.sum(truth[:, cl - 1] >= class_weights_threshold[cl - 1]))
            class_count[n][idx] = truth_sum

    return class_count


def get_total_class_count(mixdb: dict) -> List[int]:
    total_class_count = [0] * mixdb['num_classes']
    for mixture in mixdb['mixtures']:
        truth_indices = mixdb['targets'][mixture['target_file_index']]['truth_index']
        for n in range(len(truth_indices)):
            for idx, cl in enumerate(truth_indices[n]):
                total_class_count[cl - 1] += int(mixture['class_count'][n][idx])

    return total_class_count



from copy import deepcopy
from typing import List
from typing import Union

from sonusai import SonusAIError
from sonusai import logger
from sonusai.mixture import get_mixtures_from_mixid


def new_mixdb_from_mixid(mixdb: dict,
                         mixid: Union[str, List[int]]) -> dict:
    mixdb_out = deepcopy(mixdb)
    mixdb_out['mixtures'] = get_mixtures_from_mixid(mixdb_out['mixtures'], mixid)

    if not mixdb_out['mixtures']:
        logger.exception(f'Error processing mixid: {mixid}; resulted in empty list of mixtures')
        raise SonusAIError

    return mixdb_out



from copy import deepcopy

import h5py
import numpy as np
from pyaaware import ForwardTransform
from pyaaware import SED

from sonusai import SonusAIError
from sonusai import logger
from sonusai.mixture import get_class_count
from sonusai.mixture import get_class_weights_threshold
from sonusai.utils import int16_to_float


def strictly_decreasing(list_to_check: list) -> bool:
    return all(x > y for x, y in zip(list_to_check, list_to_check[1:]))


def generate_truth(mixdb: dict,
                   record: dict,
                   target: np.ndarray,
                   noise: np.ndarray,
                   compute: bool = True) -> np.ndarray:
    if not compute:
        return np.empty(0, dtype=np.single)

    truth_config = deepcopy(mixdb['targets'][record['target_file_index']]['truth_config'])
    truth_config['frame_size'] = mixdb['frame_size']
    truth_config['num_classes'] = mixdb['num_classes']
    truth_config['mutex'] = mixdb['truth_mutex']
    truth_config['target_gain'] = record['target_gain']

    indices = mixdb['targets'][record['target_file_index']]['truth_index']
    functions = mixdb['targets'][record['target_file_index']]['truth_function']

    if len(functions) != len(indices):
        logger.exception('length of truth_index does not match length of truth_function')
        raise SonusAIError

    truth = np.zeros((len(target), truth_config['num_classes']), dtype=np.single)
    for i in range(len(functions)):
        truth_config['function'] = functions[i]
        truth_config['index'] = indices[i]
        new_truth = truth_function(target=target,
                                   noise=noise,
                                   config=truth_config)
        truth = truth + new_truth

    record['class_count'] = get_class_count(
        truth_index=indices,
        truth=truth,
        class_weights_threshold=get_class_weights_threshold(mixdb))

    return truth


def truth_function(target: np.ndarray,
                   noise: np.ndarray,
                   config: dict) -> np.ndarray:
    if config['function'] == 'sed':
        if config['target_gain'] == 0:
            return np.zeros((len(target), config['num_classes']), dtype=np.single)

        if len(target) % config['frame_size'] != 0:
            logger.exception(f'Number of samples in audio is not a multiple of {config["frame_size"]}')
            raise SonusAIError

        if 'thresholds' in config.keys():
            thresholds = config['thresholds']
            if not isinstance(thresholds, list) or len(thresholds) != 3:
                logger.exception(f'Truth function SED thresholds does not contain 3 entries: {thresholds}')
                raise SonusAIError
            if not strictly_decreasing(thresholds):
                logger.exception(f'Truth function SED thresholds are not strictly decreasing: {thresholds}')
                raise SonusAIError
        else:
            thresholds = None

        fft = ForwardTransform(N=config['frame_size'] * 4, R=config['frame_size'])
        sed = SED(thresholds=thresholds,
                  index=config['index'],
                  frame_size=config['frame_size'],
                  num_classes=config['num_classes'],
                  mutex=config['mutex'])

        audio = np.int16(np.single(target) / config['target_gain'])
        truth = np.empty((0, config['num_classes']), dtype=np.single)
        for offset in range(0, len(audio), config['frame_size']):
            new_truth = sed.execute(fft.energy(int16_to_float(audio[offset:offset + config['frame_size']])))
            truth = np.vstack((truth, np.reshape(new_truth, (1, len(new_truth)))))

        truth = truth.repeat(config['frame_size'], axis=0)
        return truth

    elif config['function'] == 'file':
        file_parameters = ['file']
        for parameter in file_parameters:
            if parameter not in config.keys():
                logger.exception(f'Truth function config missing required parameter: {parameter}')
                raise SonusAIError

        with h5py.File(name=config['file'], mode='r') as f:
            truth_in = f['/truth_t'][:]

        if truth_in.ndim != 2:
            logger.exception('Truth file data is not 2 dimensions')
            raise SonusAIError

        if truth_in.shape[0] != len(target):
            logger.exception('Truth file does not contain the right amount of samples')
            raise SonusAIError

        truth = np.zeros((len(target), config['num_classes']), dtype=np.single)
        if config['target_gain'] == 0:
            return truth

        if isinstance(config['index'], list):
            if len(config['index']) != truth_in.shape[1]:
                print('Truth file does not contain the right amount of classes')
                raise SonusAIError

            truth[:, config['index']] = truth_in
        else:
            if config['index'] + truth_in.shape[1] > config['num_classes']:
                print('Truth file contains too many classes')
                raise SonusAIError

            truth[:, config['index']:config['index'] + truth_in.shape[1]] = truth_in

        return truth

    elif config['function'] == 'snr':
        truth = np.zeros((len(target), config['num_classes']), dtype=np.single)
        fft = ForwardTransform(N=config['frame_size'] * 4, R=config['frame_size'])
        for offset in range(0, len(target), config['frame_size']):
            target_energy = fft.energy(int16_to_float(target[offset:offset + config['frame_size']]))
            noise_energy = fft.energy(int16_to_float(noise[offset:offset + config['frame_size']]))
            truth[offset:offset + config['frame_size'], config['index']] = np.single(target_energy / noise_energy)

        return truth

    elif config['function'] == 'phoneme':
        # Read in .txt transcript and run a Python function to generate text grid data
        # (indicating which phonemes are active)
        # Then generate truth based on this data and put in the correct classes based on config['index']
        logger.exception('phoneme function is not supported yet')
        raise SonusAIError

    logger.exception(f'Unsupported truth function: {config["function"]}')
    raise SonusAIError



import numpy as np
from sklearn.metrics import confusion_matrix

from sonusai import SonusAIError
from sonusai import logger
from sonusai.metrics import calculate_metrics


def one_hot(truth: np.ndarray,
            predict: np.ndarray,
            pthrmode: int = 0,
            num_classes: int = -1) -> (np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray):
    # Calculates metrics from one-hot prediction and truth data where input
    # is one-hot probabilities for each class with size frames x num_classes or
    # frames x timesteps x num_classes.
    #
    # num_classes is inferred from truth dims by default (num_classes = -1).
    # Only set in case of binary 2-dimensional input frames x timesteps with timestep,
    # then must set num_classes = 1.
    #
    # returns metrics over all frames + timesteps:
    # mcm      num_classes x 2 x 2         multiclass confusion matrix count ove
    # metrics  num_classes x 10            [ACC, TPR, PPV, TNR, FPR, HITFA, F1, MCC, NT, PT, TP, FP]
    # cm, cmn: num_classes x num_classes   confusion matrix, normalized confusion matrix
    # rmse:    num_classes x 1             RMS error over all frames + timesteps

    if truth.shape != predict.shape:
        logger.exception('Shape of truth and predict are not equal')
        raise SonusAIError

    # truth, predict can be either frames x num_classes, or frames x timesteps x num_classes
    # in binary case, num_classes == 1 and dim may not exist
    if truth.ndim == 3 or (truth.ndim == 2 and num_classes == 1):  # flatten
        # frames = truth.shape[0]
        # timesteps = truth.shape[1]
        if truth.ndim == 2:
            num_classes = 1
        else:
            num_classes = truth.shape[2]

        truth = np.reshape(truth, (truth.shape[0] * truth.shape[1], truth.shape[2]))
        predict = np.reshape(predict, (predict.shape[0] * predict.shape[1], predict.shape[2]))
    else:
        # frames = truth.shape[0]
        # timesteps = 0
        if truth.ndim == 1:
            num_classes = 1
        else:
            num_classes = truth.shape[1]

    rmse = np.sqrt(np.mean(np.square(truth - predict), axis=0))

    if pthrmode == 0 and num_classes > 1:
        pthr = 0.5  # multiclass, single-label (argmax) or multilabel case default
    else:
        if num_classes == 1 and pthrmode == 0:
            pthr = 0.5  # binary case default >= 0.5 which is equiv to argmax()
        else:
            pthr = pthrmode  # any case using specified threshold

    # Convert continuous probabilities to binary via argmax() or threshold comparison
    # and create labels of int encoded (0:num_classes-1), and then equivalent one-hot
    if num_classes == 1:  # If binary
        binary_mode = True
        labels = ([i for i in range(0, 2)])  # int encoded 0,1
        plabel = np.int8(predict >= pthr)  # frames x 1, default 0.5 is equiv. to argmax()
        tlabel = np.int8(truth >= pthr)  # frames x 1
        predb = plabel
        truthb = tlabel
    else:
        binary_mode = False
        labels = ([i for i in range(0, num_classes)])  # int encoded 0,...,num_classes-1
        if pthrmode == 0:  # multiclass, use argmax (i.e. single label mutex)
            plabel = np.argmax(predict, axis=-1)  # frames x 1 labels
            tlabel = np.argmax(truth, axis=-1)  # frames x 1 labels
            predb = np.zeros(predict.shape, dtype=np.int8)  # frames x num_classes one-hot binary
            truthb = np.zeros(truth.shape, dtype=np.int8)  # frames x num_classes one-hot binary
            predb[np.arange(predb.shape[0]), plabel] = 1
            truthb[np.arange(truthb.shape[0]), tlabel] = 1
        else:  # multiclass prob threshold comparison (i.e. multi-label)
            plabel = np.int8(predict >= pthr)  # frames x num_classes multilabel one-hot decision
            tlabel = np.int8(truth >= pthr)  # frames x num_classes multilabel one-hot decision

    # Create num_classes x num_classes normalized confusion matrix
    cmn = confusion_matrix(tlabel, plabel, labels=labels, normalize='true')

    # Create num_classes x num_classes confusion matrix
    cm = confusion_matrix(tlabel, plabel, labels=labels)

    (metrics, mcm, _) = calculate_metrics(cm, binary_mode)

    return mcm, metrics, cm, cmn, rmse



from typing import List

import numpy as np

from sonusai.metrics import averages


def generate_summary(metrics: np.ndarray, target_names: List[str] = None, otheridx: int = -2, digits: int = 2) -> str:
    # Generate multi-class metric summary into string report, where
    # the simplest usage is print(generate_summary(metrics))
    #                 PPV     TPR      f1     FPR     ACC   support
    #     Class 1     0.71    0.80    0.75    0.00    0.99      44
    #     Class 2     0.90    0.76    0.82    0.00    0.99     128
    #     Class 3     0.86    0.82    0.84    0.04    0.93     789
    #     Other       0.94    0.96    0.95    0.18    0.92    2807
    #
    #   micro-avg                     0.92    0.027           3768
    #   macro avg     0.85    0.83    0.84    0.05    0.96    3768
    #   micro-avgwo
    #
    # metrics is num_classesx12 data from confusion_matrix() that calcs stats on raw data.
    # otheridx is used to declare which category is the "other" class for
    # single label multi-class classification, and reports will use this,
    # otheridx == -1 specifies the other class is not present and
    # micro-avgwo and macro-avgwo metrics are not printed.
    # otheridx == -2 is special case declaring other class in last class index = num_classes
    # For binary classification otheridx is ignored.
    num_classes = np.shape(metrics)[0]  # num classes, assumes metrics is from confusion_matrix()
    if num_classes == 1:
        otheridx = -1

    if otheridx == -2:
        otheridx = num_classes - 1

    if target_names is None or len(target_names) != num_classes:
        # target_names = ['%s' % l for l in labels]
        target_names = ([f'Class {i}' for i in range(1, num_classes + 1)])
        if otheridx > -1:
            target_names[otheridx] = 'Other'

    # format from sklearn,  headers = ["precision", "recall", "f1-score", "support"]
    headers = ["PPV", "TPR", "F1", "FPR", "ACC", "Support"]
    p = metrics[:, 2]
    r = metrics[:, 1]
    fpr = metrics[:, 4]
    f1 = metrics[:, 6]
    acc = metrics[:, 0]
    s = metrics[:, 9].astype(int)
    rows = zip(target_names, p, r, f1, fpr, acc, s)

    longest_last_line_heading = 'weighted avg'
    name_width = max(len(cn) for cn in target_names)
    width = max(name_width, len(longest_last_line_heading), digits)
    head_fmt = '{:>{width}s} ' + ' {:>7}' * len(headers)
    report = head_fmt.format('', *headers, width=width)
    report += '\n'
    row_fmt = '{:>{width}s} ' + ' {:>7.{digits}f}' * 5 + ' {:>7}\n'
    for row in rows:
        report += row_fmt.format(*row, width=width, digits=digits)
    report += '\n'

    if num_classes > 1:  # Binary does not require average
        # compute averages, all options are average_options = ('micro', 'macro', 'weighted', 'samples')
        if otheridx > -1:
            average_options = ('micro-avg', 'macro-avg', 'micro-avgwo', 'macro-avgwo')
            tmp = np.delete(metrics, otheridx, 0)
            mavgwo = averages(tmp)  # avg without other class
        else:
            average_options = ('micro-avg', 'macro-avg')

        mavg = averages(metrics)  # [PPV, TPR, F1, FPR, ACC, TPSUM]
        for average in average_options:
            line_heading = average

            # compute averages with specified averaging method
            if average == 'macro-avg':
                idx = 0
                # [np.mean(p), np.mean(r), np.mean(f1), np.mean(fpr), np.mean(acc), np.sum(s)]
                avg = [mavg[idx, 0], mavg[idx, 1], mavg[idx, 2],
                       mavg[idx, 3], mavg[idx, 4], mavg[idx, 5].astype(int)]
            elif average == 'micro-avg':
                # micro-mode : micro-F1 = micro-precision = micro-recall = accuracy
                idx = 1
                avg = [mavg[idx, 0], mavg[idx, 1], mavg[idx, 2],
                       mavg[idx, 3], mavg[idx, 4], mavg[idx, 5].astype(int)]
            elif average == 'macro-avgwo':
                idx = 0
                # [np.mean(p), np.mean(r), np.mean(f1), np.mean(fpr), np.mean(acc), np.sum(s)]
                avg = [mavgwo[idx, 0], mavgwo[idx, 1], mavgwo[idx, 2],
                       mavgwo[idx, 3], mavgwo[idx, 4], mavgwo[idx, 5].astype(int)]
            elif average == 'micro-avgwo':
                # micro-mode : micro-F1 = micro-precision = micro-recall = accuracy
                idx = 1
                avg = [mavgwo[idx, 0], mavgwo[idx, 1], mavgwo[idx, 2],
                       mavgwo[idx, 3], mavgwo[idx, 4], mavgwo[idx, 5].astype(int)]
            elif average == 'weighted':
                avg = mavg[2,]  # TBD

            if line_heading.startswith('micro'):
                row_fmt_accuracy = '{:>{width}s} ' + ' {:>7.{digits}}' * 4 + ' {:>7.{digits}s}' + ' {:>7}\n'
                report += row_fmt_accuracy.format(line_heading, '', '', *avg[2:4], '', avg[5], width=width,
                                                  digits=digits)
            else:
                report += row_fmt.format(line_heading, *avg,
                                         width=width, digits=digits)

    return report



import h5py
import numpy as np


def read_mixture_data(filename: str) -> (np.ndarray, np.ndarray, np.ndarray):
    if not filename:
        return None, None, None

    with h5py.File(name=filename, mode='r') as f:
        return f['mixture'][:], f['target'][:], f['noise'][:]



from typing import List
from typing import Union


def get_mixtures_from_mixid(mix_in: list, mixid: Union[str, List[int]]) -> list:
    if isinstance(mixid, str):
        try:
            mix_out = eval(f'mix_in[{mixid}]')
            if not isinstance(mix_out, list):
                mix_out = [mix_out]
            return mix_out
        except NameError:
            return []

    elif isinstance(mixid, list) and all(isinstance(x, int) and x < len(mix_in) for x in mixid):
        return [mix_in[i] for i in mixid]

    return []



"""predict

usage: predict [-hvnlas] (-m MODEL) (-i INPUT) [-o OUTPUT] [-e EXPECT] [-r RESET]

options:
   -h, --help
   -v, --verbose                Be verbose.
   -n, --no_progress            Do not show the progress bar.
   -m MODEL, --model MODEL      Trained model file.
   -i INPUT, --input INPUT      Input feature file.
   -o OUTPUT, --output OUTPUT   Optional output directory.
   -e EXPECT, --expect EXPECT   Optional expected results file.
   -l, --legacy                 Use legacy AawareRT.
   -a, --accelerate             Use acceleration if possible (only in legacy mode).
   -r RESET, --reset RESET      Reset model after RESET frames (only in legacy mode).
   -s, --sonusai_ep             Use the ONNX runtime SonusAI execution provider (only in non-legacy mode).

The predict command runs predictions on a trained model using genft generated data
and optionally compares the results to expected results.

"""
from datetime import datetime
from os import mkdir
from subprocess import run

from docopt import docopt

import sonusai
from sonusai import SonusAIError
from sonusai import create_file_handler
from sonusai import initial_log_messages
from sonusai import logger
from sonusai import update_console_handler
from sonusai.utils import trim_docstring


def predict(model: str,
            feature: str,
            output: str = None,
            expect: str = None,
            legacy: bool = False,
            accelerate: bool = False,
            reset: int = None,
            sonusai_ep: bool = True,
            verbose: bool = False,
            show_progress: bool = False):
    update_console_handler(verbose)
    initial_log_messages('predict')

    # create output directory
    if output is None:
        output = f'predict-{datetime.now():%Y%m%d-%H%M%S}'

    try:
        mkdir(output)
    except OSError as e:
        logger.exception(f'Could not create directory, {output}: {e}')
        raise SonusAIError

    log_name = output + '/predict.log'
    create_file_handler(log_name)

    logger.info(f'Writing results to {output}')

    # create command
    predict_type = ''
    if legacy:
        predict_type = '-art'

    command = ['sonusai-helper-predict' + predict_type, '-m', model, '-i', feature, '-o', output]

    if expect is not None:
        command.extend(['-e', expect])

    if legacy and accelerate:
        command.append('-a')

    if legacy and reset is not None:
        command.extend(['-r', reset])

    if not legacy and sonusai_ep:
        command.extend(['-s'])

    if verbose:
        command.append('-v')

    if not show_progress:
        command.extend(['-n'])

    try:
        run(command, check=True)
    except Exception as e:
        logger.exception(f'Error running predict: {e}')
        raise SonusAIError


def main():
    try:
        args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

        predict(model=args['--model'],
                feature=args['--input'],
                output=args['--output'],
                expect=args['--expect'],
                legacy=args['--legacy'],
                accelerate=args['--accelerate'],
                reset=args['--reset'],
                sonusai_ep=args['--sonusai_ep'],
                verbose=args['--verbose'],
                show_progress=not args['--no_progress'])

    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)


if __name__ == '__main__':
    main()



"""genmixdb

usage: genmixdb [-hv] CONFIG...

options:
   -h, --help
   -v, --verbose    Be verbose.

Create mixture database data for training and evaluation.

genmixdb creates a database of training and evaluation feature and truth data generation information. It allows
the choice of audio neural-network feature types that are supported by the Aaware real-time front-end and truth
data that is synchronized frame-by-frame with the feature data.

Here are some examples:

#### Adding target data
Suppose you have an audio file which is an example, or target, of what you want to
recognize or detect. Of course, for training a NN you also need truth data for that
file (also called labels). If you don't already have it, genmixdb can create truth using
energy-based sound detection on each frame of the feature data. You can also select
different feature types. Here's an example:

genmixdb target_gfr32ts2.yml

where target_gfr32ts2.yml contains:
---
targets:
  - data/target.wav

feature: gfr32ts2

target_augmentations:
  -
    normalize: -3.5
...

The mixture database is written to a JSON file that inherits the base name of the config file.

#### Target data mix with noise and augmentation

genmixdb mix_gfr32ts2.yml

where mix_gfr32ts2.yml contains:
---
targets:
  - data/target.wav

noises:
  - data/noise.wav

target_count: 5

feature: gfr32ts2

output: data/my_mix.h5

target_augmentations:
  -
    normalize: -3.5
    pitch: [-3, 0, 3]
    tempo: [0.8, 1, 1.2]
    snr: 20

noise_augmentations:
  -
    normalize: -3.5
...

In this example a time-domain mixture is created and feature data is calculated as
specified by 'feature: gfr32ts2'. Various feature types are available which vary in
spectral and temporal resolution (4 ms or higher), and other feature algorithm
parameters. The total feature size, dimension, and #frames for mixture is reported
in the log file (the log file name is derived from the output file base name; in this
case it would be mix_gfr32ts2.log).

Truth (labels) can be automatically created per feature output frame based on sound
energy detection. By default, these are appended to the feature data in a single HDF5
output file. By default, truth/label generation is turned on with default algorithm
and threshold levels (see truth section) and a single class, i.e., detecting a single
type of sound. The truth format is a single float per class representing the
probability of activity/presence, and multi-class truth/labels are possible by
specifying the number of classes and either a scalar index or a vector of indices in
which to put the truth result. For example, 'num_class: 3' and  'truth_index: 2' adds
a 1x3 vector to the feature data with truth put in index 2 (others would be 0) for
data/target.wav being an audio clip from sound type of class 2.

The mixture is created with potential data augmentation functions in the following way:
1. apply noise augmentation rule
2. apply target augmentation rule
3. adjust noise gain for specific SNR
4. add augmented noise to augmented target

The mixture length is the target length by default, and the noise signal is repeated
if it is shorter, or trimmed if longer. If 'target_count: <count>' is specified, then
the target audio is concatenated <count> times for each augmentation rule.
(See the Augmentation section for details on augmentation rules.)

#### Target and noise using path lists

Target and noise audio is specified as a list containing text files, audio files, and
file globs. Text files are processed with items on each line where each item can be a
text file, an audio file, or a file glob. Each item will be searched for audio files
which can be WAV, MP3, FLAC, AIFF, or OGG format with any sample rate, bit depth, or
channel count. All audio files will be converted to 16 kHz, 16-bit, single channel
format before processing. For example,

genmixdb dog-bark.yml

where dog-bark.yml contains:
---
targets:
  - slib/dog-outside/*.wav
  - slib/dog-inside/*.wav

will find all .wav files in the specified directories and process them as targets.

"""
import json
import re
import time
from copy import deepcopy
from glob import glob
from numbers import Number
from os import listdir
from os.path import dirname
from os.path import isabs
from os.path import isdir
from os.path import join
from os.path import splitext
from random import seed
from random import uniform

import numpy as np
import sox
import yaml
from docopt import docopt
from pyaaware import FeatureGenerator
from tqdm import tqdm

import sonusai
from sonusai import SonusAIError
from sonusai import create_file_handler
from sonusai import initial_log_messages
from sonusai import logger
from sonusai import update_console_handler
from sonusai.mixture import apply_augmentation
from sonusai.mixture import estimate_audio_length
from sonusai.mixture import generate_truth
from sonusai.mixture import get_class_weights_threshold
from sonusai.mixture import get_config_from_file
from sonusai.mixture import get_next_noise
from sonusai.mixture import get_total_class_count
from sonusai.mixture import read_audio
from sonusai.utils import expandvars
from sonusai.utils import human_readable_size
from sonusai.utils import p_map
from sonusai.utils import seconds_to_hms
from sonusai.utils import trim_docstring

genmixdb_global = dict()


def genmixdb(config: dict, logging: bool = True, show_progress: bool = False) -> dict:
    required_keys = [
        'class_labels',
        'class_weights_threshold',
        'dither',
        'feature',
        'frame_size',
        'noises',
        'noise_augmentations',
        'num_classes',
        'seed',
        'target_augmentations',
        'targets',
        'truth_config',
        'truth_function',
        'truth_index',
        'truth_mode',
        'truth_reduction_function',
    ]
    for key in required_keys:
        if key not in config.keys():
            logger.exception(f'Missing {key} in config')
            raise SonusAIError

    seed(config['seed'])

    if logging:
        logger.debug(f'Seed: {config["seed"]}')
        logger.debug('Configuration:')
        logger.debug(yaml.dump(config))

    if logging:
        logger.info('Collecting targets')
    targets = get_input_files(config['targets'], cfg={'truth_index':    config['truth_index'],
                                                      'truth_function': config['truth_function'],
                                                      'truth_config':   config['truth_config']})
    if len(targets) == 0:
        logger.exception('Canceled due to no targets')
        raise SonusAIError

    if logging:
        logger.debug('List of targets:')
        logger.debug(yaml.dump([sub['name'] for sub in targets], default_flow_style=False))

    if logging:
        logger.info('Collecting noises')
    noises = get_input_files(config['noises'])
    if len(noises) == 0:
        if logging:
            logger.warning(f'Did not find any noises; using default noise: {sonusai.mixture.default_noise}')
        noises = get_input_files([sonusai.mixture.default_noise])
    if logging:
        logger.debug('List of noises:')
        logger.debug(yaml.dump([sub['name'] for sub in noises], default_flow_style=False))

    if logging:
        logger.info('Collecting target augmentations')
    target_augmentations = get_augmentations(config['target_augmentations'], target=True)
    expanded_target_augmentations = ''
    for augmentation in target_augmentations:
        expanded_target_augmentations += f'- {augmentation}\n'
    if logging:
        logger.debug('Expanded list of target augmentations:')
        logger.debug(expanded_target_augmentations)

    if logging:
        logger.info('Collecting noise augmentations')
    noise_augmentations = get_augmentations(config['noise_augmentations'], target=False)
    expanded_noise_augmentations = ''
    for augmentation in noise_augmentations:
        expanded_noise_augmentations += f'- {augmentation}\n'
    if logging:
        logger.debug('Expanded list of noise augmentations:')
        logger.debug(expanded_noise_augmentations)

    noise_sets = len(noises) * len(noise_augmentations)
    target_sets = len(targets) * len(target_augmentations)
    total_mixtures = noise_sets * target_sets
    if logging:
        logger.info('')
        logger.info(f'Found {total_mixtures} mixtures to process')

    class_weights_threshold = get_class_weights_threshold(config)

    if config['truth_mode'] not in ['normal', 'mutex']:
        logger.exception(f'invalid truth_mode: {config["truth_mode"]}')
        raise SonusAIError

    truth_mutex = config['truth_mode'] == 'mutex'
    if truth_mutex:
        max_class = config['num_classes'] - 1
    else:
        max_class = config['num_classes']

    fg = FeatureGenerator(feature_mode=config['feature'],
                          frame_size=config['frame_size'],
                          num_classes=config['num_classes'],
                          truth_mutex=truth_mutex)

    num_bands = fg.num_bands
    stride = fg.stride
    step = fg.step
    decimation = fg.decimation

    transform_frame_ms = float(config['frame_size']) / float(sonusai.mixture.sample_rate / 1000)
    feature_ms = transform_frame_ms * decimation * stride
    feature_step_ms = transform_frame_ms * decimation * step
    feature_samples = config['frame_size'] * decimation * stride
    feature_step_samples = config['frame_size'] * decimation * step

    total_duration = 0
    for duration in [sub['duration'] for sub in targets]:
        for augmentation in target_augmentations:
            length = int(duration * sonusai.mixture.sample_rate)
            if 'tempo' in augmentation:
                length /= augmentation['tempo']
            if length % feature_step_samples:
                length += feature_step_samples - int(length % feature_step_samples)
            total_duration += float(length) / sonusai.mixture.sample_rate
    total_duration *= noise_sets

    total_samples = total_duration * sonusai.mixture.sample_rate

    mixture_bytes = total_samples * sonusai.mixture.bit_depth / 8
    truth_t_bytes = total_samples * config['num_classes'] * sonusai.mixture.float_bytes
    feature_bytes = total_samples / feature_step_samples * stride * num_bands * sonusai.mixture.float_bytes
    truth_f_bytes = total_samples / feature_step_samples * config['num_classes'] * sonusai.mixture.float_bytes

    if logging:
        logger.info('')
        logger.info(f'Estimated duration:   {seconds_to_hms(seconds=total_duration)}')
        logger.info('Estimated sizes:')
        logger.info(f' mixture:             {human_readable_size(mixture_bytes, 1)}')
        logger.info(f' truth_t:             {human_readable_size(truth_t_bytes, 1)}')
        logger.info(f' feature:             {human_readable_size(feature_bytes, 1)}')
        logger.info(f' truth_f:             {human_readable_size(truth_f_bytes, 1)}')
        logger.info(f'Feature shape:        {stride} x {num_bands} ({stride * num_bands} total params)')
        logger.info(f'Feature samples:      {feature_samples} samples ({feature_ms} ms)')
        logger.info(f'Feature step samples: {feature_step_samples} samples ({feature_step_ms} ms)')

    mixdb = {
        'class_count':              list(),
        'class_labels':             config['class_labels'],
        'class_weights_threshold':  list(class_weights_threshold),
        'dither':                   config['dither'],
        'feature':                  config['feature'],
        'feature_samples':          feature_samples,
        'feature_step_samples':     feature_step_samples,
        'frame_size':               config['frame_size'],
        'mixtures':                 list(),
        'noise_augmentations':      noise_augmentations,
        'noises':                   noises,
        'num_classes':              config['num_classes'],
        'target_augmentations':     target_augmentations,
        'targets':                  targets,
        'truth_mutex':              truth_mutex,
        'truth_reduction_function': config['truth_reduction_function'],
    }

    # Read in all audio data beforehand to avoid reading it multiple times in the loop
    if logging:
        logger.debug('Reading noise audio')
    raw_noise_audio = list()
    for noise in noises:
        raw_noise_audio.append(read_audio(name=noise['name'], dither=config['dither']))

    if logging:
        logger.debug('Reading target audio')
    raw_target_audio = list()
    for target in targets:
        raw_target_audio.append(read_audio(name=target['name'], dither=config['dither']))

    # First pass to get indices and offsets
    mixtures = [[] for _ in range(noise_sets)]
    n_id = 0
    for noise_index, noise in enumerate(noises):
        for noise_augmentation_index, noise_augmentation in enumerate(noise_augmentations):
            mixtures[n_id] = [[] for _ in range(target_sets)]
            t_id = 0
            noise_offset = 0
            noise_length = estimate_audio_length(audio_in=raw_noise_audio[noise_index],
                                                 augmentation=noise_augmentation,
                                                 length_common_denominator=1)
            for target_index, target in enumerate(targets):
                for target_augmentation_index, target_augmentation in enumerate(target_augmentations):
                    if not isinstance(target['truth_function'], list):
                        target['truth_function'] = [target['truth_function']]

                    if not isinstance(target['truth_index'], list):
                        target['truth_index'] = [target['truth_index']]

                    if len(target['truth_function']) != len(target['truth_index']):
                        logger.exception('length of truth_index does not match length of truth_function')
                        raise SonusAIError

                    for i in range(len(target['truth_function'])):
                        if not isinstance(target['truth_index'][i], list):
                            target['truth_index'][i] = [target['truth_index'][i]]
                        if any(idx > max_class for idx in target['truth_index'][i]):
                            logger.exception('invalid truth_index')
                            raise SonusAIError

                    mixtures[n_id][t_id] = {'target_file_index':         target_index,
                                            'noise_file_index':          noise_index,
                                            'noise_offset':              noise_offset,
                                            'target_augmentation_index': target_augmentation_index,
                                            'noise_augmentation_index':  noise_augmentation_index}
                    t_id += 1

                    target_length = estimate_audio_length(audio_in=raw_target_audio[target_index],
                                                          augmentation=target_augmentation,
                                                          length_common_denominator=feature_step_samples)
                    noise_offset = int((noise_offset + target_length) % noise_length)

            n_id += 1

    genmixdb_global['mixdb'] = mixdb
    genmixdb_global['config'] = config
    genmixdb_global['target_augmentations'] = target_augmentations
    genmixdb_global['raw_target_audio'] = raw_target_audio
    genmixdb_global['feature_step_samples'] = feature_step_samples

    progress = tqdm(total=total_mixtures, desc='genmixdb', disable=not show_progress)

    # Second pass to fill in the details
    for n_id in range(len(mixtures)):
        genmixdb_global['augmented_noise_audio'] = apply_augmentation(
            audio_in=raw_noise_audio[mixtures[n_id][0]['noise_file_index']],
            augmentation=noise_augmentations[mixtures[n_id][0]['noise_augmentation_index']],
            length_common_denominator=1,
            dither=config['dither'])
        mixtures[n_id] = p_map(process_mixture, mixtures[n_id], progress=progress)

    progress.close()

    # Flatten mixtures
    mixdb['mixtures'] = [item for sublist in mixtures for item in sublist]
    mixdb['class_count'] = get_total_class_count(mixdb)

    total_samples = sum([sub['samples'] for sub in mixdb['mixtures']])
    total_duration = total_samples / sonusai.mixture.sample_rate
    mixture_bytes = total_samples * sonusai.mixture.bit_depth / 8
    truth_t_bytes = total_samples * config['num_classes'] * 4
    feature_bytes = total_samples / feature_step_samples * stride * num_bands * 4
    truth_f_bytes = total_samples / feature_step_samples * config['num_classes'] * 4
    if logging:
        logger.info('')
        logger.info(f'Actual duration: {seconds_to_hms(seconds=total_duration)}')
        logger.info('Actual sizes:')
        logger.info(f' mixture:        {human_readable_size(mixture_bytes, 1)}')
        logger.info(f' truth_t:        {human_readable_size(truth_t_bytes, 1)}')
        logger.info(f' feature:        {human_readable_size(feature_bytes, 1)}')
        logger.info(f' truth_f:        {human_readable_size(truth_f_bytes, 1)}')

    return mixdb


def get_input_files(records, cfg=None):
    if cfg is None:
        cfg = dict()

    files = list()
    for record in records:
        append_input_files(files, record, cfg)
    return files


def append_input_files(files: list, in_record, cfg: dict, tokens=None):
    if tokens is None:
        tokens = dict()
    in_name = in_record

    if isinstance(in_record, dict):
        if 'target_name' in in_record.keys():
            in_name = in_record['target_name']
        else:
            logger.exception('Target list contained record without \'target_name\' key')
            raise SonusAIError

        if 'truth_index' in in_record.keys():
            cfg['truth_index'] = in_record['truth_index']

        if 'truth_function' in in_record.keys():
            cfg['truth_function'] = in_record['truth_function']

        if 'truth_config' in in_record.keys():
            cfg['truth_config'] = in_record['truth_config']

    (in_name, new_tokens) = expandvars(in_name)
    tokens.update(new_tokens)
    names = glob(in_name)
    if not names:
        logger.exception(f'Could not find {in_name}. Make sure path exists')
        raise SonusAIError
    for name in names:
        ext = splitext(name)[1].lower()
        dir_name = dirname(name)
        if isdir(name):
            for file in listdir(name):
                child = file
                if not isabs(child):
                    child = join(dir_name, child)
                append_input_files(files, child, cfg, tokens)
        else:
            try:
                if ext == '.txt':
                    with open(name, mode='r') as txt_file:
                        for line in txt_file:
                            # strip comments
                            child = line.partition('#')[0]
                            child = child.rstrip()
                            if child:
                                (child, new_tokens) = expandvars(child)
                                tokens.update(new_tokens)
                                if not isabs(child):
                                    child = join(dir_name, child)
                                append_input_files(files, child, cfg, tokens)
                elif ext == '.yml':
                    try:
                        with open(name, mode='r') as yml_file:
                            yml_config = yaml.safe_load(yml_file)

                        if 'targets' in yml_config:
                            for record in yml_config['targets']:
                                append_input_files(files, record, cfg, tokens)
                    except Exception as e:
                        logger.exception(f'Error processing {name}: {e}')
                        raise SonusAIError
                else:
                    sox.file_info.validate_input_file(name)
                    duration = sox.file_info.duration(name)
                    for key, value in tokens.items():
                        name = name.replace(value, f'${key}')
                    entry = {'name': name, 'duration': duration}
                    entry.update(cfg)
                    if 'truth_function' in entry.keys() and entry['truth_function'] == 'file':
                        entry['truth_config']['file'] = splitext(name)[0] + '.h5'
                    files.append(entry)
            except SonusAIError:
                raise
            except Exception as e:
                logger.exception(f'Error processing {name}: {e}')
                raise SonusAIError


def get_augmentations(rules: list, target: bool) -> list:
    augmentations = list()
    for rule in rules:
        expand_augmentations(augmentations, rule, target)

    augmentations = rand_augmentations(augmentations)
    return augmentations


def expand_augmentations(augmentations: list, rule: dict, target: bool):
    # replace old 'eq' rule with new 'eq1' rule to allow both for backward compatibility
    rule = {'eq1' if key == 'eq' else key: value for key, value in rule.items()}

    for key in rule:
        if key not in sonusai.mixture.valid_augmentations:
            logger.exception(f'Invalid augmentation: {key}')
            raise SonusAIError

        if key in ['eq1', 'eq2', 'eq3']:
            # eq must be a list of length 3 or a list of length 3 lists
            valid = True
            multiple = False
            if isinstance(rule[key], list):
                if any(isinstance(el, list) for el in rule[key]):
                    multiple = True
                    for value in rule[key]:
                        if not isinstance(value, list) or len(value) != 3:
                            valid = False
                else:
                    if len(rule[key]) != 3:
                        valid = False
            else:
                valid = False

            if not valid:
                logger.exception(f'Invalid augmentation value for {key}: {rule[key]}')
                raise SonusAIError

            if multiple:
                for value in rule[key]:
                    expanded_rule = deepcopy(rule)
                    expanded_rule[key] = deepcopy(value)
                    expand_augmentations(augmentations, expanded_rule, target)
                return

        elif key == 'count':
            pass

        else:
            if isinstance(rule[key], list):
                for value in rule[key]:
                    if isinstance(value, list):
                        logger.exception(f'Invalid augmentation value for {key}: {rule[key]}')
                        raise SonusAIError
                    expanded_rule = deepcopy(rule)
                    expanded_rule[key] = deepcopy(value)
                    expand_augmentations(augmentations, expanded_rule, target)
                return
            elif not isinstance(rule[key], Number):
                if not rule[key].startswith('rand'):
                    logger.exception(f'Invalid augmentation value for {key}: {rule[key]}')
                    raise SonusAIError

    # Set default snr for target if needed
    if target and 'snr' not in rule:
        rule['snr'] = sonusai.mixture.default_snr

    # Remove snr rule from noise if needed
    if not target and 'snr' in rule:
        del rule['snr']

    augmentations.append(rule)


def rand_repl(m):
    return f'{uniform(float(m.group(1)), float(m.group(4))):.2f}'


def rand_augmentations(in_rules):
    out_rules = list()
    for rule in in_rules:
        # do any keys contain rand?
        has_rand = False
        for key in rule:
            if 'rand' in str(rule[key]):
                has_rand = True
                break

        if has_rand:
            count = 1
            if rule['count'] is not None:
                count = rule['count']
                del rule['count']
            for i in range(count):
                new_rule = deepcopy(rule)
                for key in new_rule:
                    new_rule[key] = eval(re.sub(sonusai.mixture.rand_pattern, rand_repl, str(new_rule[key])))

                    # convert eq values from strings to numbers
                    if key in ['eq1', 'eq2', 'eq3']:
                        for n in range(3):
                            if isinstance(new_rule[key][n], str):
                                new_rule[key][n] = eval(new_rule[key][n])
                out_rules.append(new_rule)
        else:
            out_rules.append(rule)
    return out_rules


def get_base_name(config: dict, config_name: str) -> str:
    try:
        config_base = splitext(config_name)[0]
        name = str(splitext(config['output'])[0])
        name = name.replace('${config}', config_base)
        return name
    except Exception as e:
        logger.exception(f'Error getting genmixdb base name: {e}')
        raise SonusAIError


def process_mixture(mixture: dict) -> dict:
    target_index = mixture['target_file_index']
    noise_offset = mixture['noise_offset']
    target_augmentation_index = mixture['target_augmentation_index']

    target_augmentation = genmixdb_global['target_augmentations'][target_augmentation_index]

    augmented_target_audio = apply_augmentation(audio_in=genmixdb_global['raw_target_audio'][target_index],
                                                augmentation=target_augmentation,
                                                length_common_denominator=genmixdb_global['feature_step_samples'],
                                                dither=genmixdb_global['config']['dither'])

    noise_segment, _ = get_next_noise(offset_in=noise_offset,
                                      length=len(augmented_target_audio),
                                      audio_in=genmixdb_global['augmented_noise_audio'])

    # target_gain is the sum of 'gain', noise overflow adjustment, and mixture
    # overflow adjustment. It is used to return the target audio to its normalized
    # level when calculating truth (unless target gain is set to zero, in which
    # case the truth will be all zeros).
    if 'gain' in target_augmentation:
        target_gain = 10 ** (target_augmentation['gain'] / 20)
    else:
        target_gain = 1

    if 'snr' in target_augmentation:
        if target_augmentation['snr'] < -96:
            # Special case for zeroing out target data
            mixture['target_snr_gain'] = 0
            mixture['noise_snr_gain'] = 1
            target_gain = 0
        else:
            target_energy = np.mean(np.square(np.single(augmented_target_audio)))
            noise_energy = np.mean(np.square(np.single(noise_segment)))
            noise_gain = np.sqrt(target_energy / noise_energy) / 10 ** (
                    target_augmentation['snr'] / 20)

            # Check for noise_gain > 1 to avoid clipping
            if noise_gain > 1:
                mixture['target_snr_gain'] = 1 / noise_gain
                mixture['noise_snr_gain'] = 1
                target_gain = target_gain / noise_gain
            else:
                mixture['target_snr_gain'] = 1
                mixture['noise_snr_gain'] = noise_gain

    # Check for clipping in mixture
    gain_adjusted_target_audio = np.single(augmented_target_audio) * mixture['target_snr_gain']
    gain_adjusted_noise_audio = np.single(noise_segment) * mixture['noise_snr_gain']
    mixture_audio = gain_adjusted_target_audio + gain_adjusted_noise_audio

    if any(abs(mixture_audio) >= 32768):
        # Clipping occurred; lower gains to bring audio within int16 bounds
        gain_adjustment = 32760 / max(abs(mixture_audio))
        mixture['target_snr_gain'] *= gain_adjustment
        mixture['noise_snr_gain'] *= gain_adjustment
        target_gain *= gain_adjustment

    mixture['samples'] = len(augmented_target_audio)
    mixture['target_gain'] = target_gain

    gain_adjusted_target_audio = np.int16(np.single(augmented_target_audio) * mixture['target_snr_gain'])
    gain_adjusted_noise_audio = np.int16(np.single(noise_segment) * mixture['noise_snr_gain'])
    generate_truth(mixdb=genmixdb_global['mixdb'],
                   record=mixture,
                   target=gain_adjusted_target_audio,
                   noise=gain_adjusted_noise_audio)

    return mixture


def main():
    try:
        args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

        verbose = args['--verbose']

        for config_file in args['CONFIG']:
            start_time = time.monotonic()
            logger.info(f'Creating mixture database for {config_file}')
            config = get_config_from_file(config_file)

            base_name = get_base_name(config, config_file)
            output_name = base_name + '.json'

            log_name = base_name + '.log'
            create_file_handler(log_name)
            update_console_handler(verbose)
            initial_log_messages('genmixdb')

            mixdb = genmixdb(config=config, show_progress=True)

            with open(output_name, mode='w') as file:
                json.dump(mixdb, file, indent=2)
                logger.info(f'Wrote mixture database for {config_file} to {output_name}')

            end_time = time.monotonic()
            logger.info(f'Completed in {seconds_to_hms(seconds=end_time - start_time)}')

    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)


if __name__ == '__main__':
    main()



import numpy as np

from sonusai import SonusAIError
from sonusai import logger
from sonusai.mixture import apply_augmentation
from sonusai.mixture import generate_segsnr
from sonusai.mixture import generate_truth
from sonusai.mixture import get_next_noise
from sonusai.mixture import get_noise_audio_from_db
from sonusai.mixture import get_target_audio_from_db


def get_target_noise_audio(mixdb: dict,
                           record: dict,
                           target_audios: list,
                           noise_audios: list) -> (np.ndarray, np.ndarray):
    if record['samples'] % mixdb['frame_size'] != 0:
        logger.exception(f'Number of samples in record is not a multiple of {mixdb["frame_size"]}')
        raise SonusAIError
    target_file_index = record['target_file_index']
    target_augmentation = mixdb['target_augmentations'][record['target_augmentation_index']]
    target_audio = apply_augmentation(audio_in=get_target_audio_from_db(target_audios, target_file_index),
                                      augmentation=target_augmentation,
                                      length_common_denominator=mixdb['feature_step_samples'],
                                      dither=mixdb['dither'])
    if len(target_audio) != record['samples']:
        logger.exception('Number of samples in target does not match database')
        raise SonusAIError
    noise_file_index = record['noise_file_index']
    noise_augmentation_index = record['noise_augmentation_index']
    noise_audio, _ = get_next_noise(offset_in=record['noise_offset'],
                                    length=record['samples'],
                                    audio_in=get_noise_audio_from_db(noise_audios,
                                                                     noise_file_index,
                                                                     noise_augmentation_index))

    target_audio = np.array(np.single(target_audio) * record['target_snr_gain'], dtype=np.int16)
    noise_audio = np.array(np.single(noise_audio) * record['noise_snr_gain'], dtype=np.int16)

    return target_audio, noise_audio


def get_audio_and_truth_t(mixdb: dict,
                          record: dict,
                          target_audios: list,
                          noise_audios: list,
                          compute_truth: bool = True,
                          compute_segsnr: bool = False,
                          frame_based_segnsr: bool = False) -> (np.ndarray, np.ndarray, np.ndarray, np.ndarray):
    target, noise = get_target_noise_audio(mixdb=mixdb,
                                           record=record,
                                           target_audios=target_audios,
                                           noise_audios=noise_audios)

    mixture = np.array(target + noise, dtype=np.int16)

    truth_t = generate_truth(mixdb=mixdb,
                             record=record,
                             target=target,
                             noise=noise,
                             compute=compute_truth)

    segsnr = generate_segsnr(mixdb=mixdb,
                             record=record,
                             target=target,
                             noise=noise,
                             compute=compute_segsnr,
                             frame_based=frame_based_segnsr)

    return mixture, truth_t, target, noise, segsnr


def set_mixture_offsets(mixdb: dict) -> None:
    i_sample_offset = 0
    i_frame_offset = 0
    o_frame_offset = 0
    for record in mixdb['mixtures']:
        record['i_sample_offset'] = i_sample_offset
        record['i_frame_offset'] = i_frame_offset
        record['o_frame_offset'] = o_frame_offset

        i_sample_offset += record['samples']
        i_frame_offset += record['samples'] // mixdb['frame_size']
        o_frame_offset += record['samples'] // mixdb['feature_step_samples']



# SonusAI mixture utilities

import os
import re

import sonusai

valid_augmentations = ['normalize', 'gain', 'pitch', 'tempo', 'eq1', 'eq2', 'eq3', 'lpf', 'snr', 'count']
rand_pattern = re.compile(r'rand\(([-+]?([0-9]+(\.[0-9]*)?|\.[0-9]+)),\s*([-+]?([0-9]+(\.[0-9]*)?|\.[0-9]+))\)')
sample_rate = 16000
bit_depth = 16
channel_count = 1
sample_bytes = int(bit_depth / 8)
float_bytes = 4
excessive_size = 6 * 1024 * 1024 * 1024

default_noise = os.path.join(sonusai.basedir, 'data', 'whitenoise.wav')
default_config = os.path.join(sonusai.basedir, 'data', 'genmixdb.yml')
default_snr = 80

from sonusai.mixture.apply_augmentation import apply_augmentation  # noqa: E402
from sonusai.mixture.apply_augmentation import read_audio  # noqa: E402
from sonusai.mixture.audio_db import build_noise_audio_db  # noqa: E402
from sonusai.mixture.audio_db import build_target_audio_db  # noqa: E402
from sonusai.mixture.audio_db import get_noise_audio_from_db  # noqa: E402
from sonusai.mixture.audio_db import get_target_audio_from_db  # noqa: E402
from sonusai.mixture.audio_db import audio_files_exist  # noqa: E402
from sonusai.mixture.get_class_weights_threshold import get_class_weights_threshold  # noqa: E402
from sonusai.mixture.get_config_from_file import get_config_from_file  # noqa: E402
from sonusai.mixture.get_next_noise import get_next_noise  # noqa: E402
from sonusai.mixture.get_offsets import get_offsets  # noqa: E402
from sonusai.mixture.class_count import get_class_count  # noqa: E402
from sonusai.mixture.class_count import get_total_class_count  # noqa: E402
from sonusai.mixture.get_mixtures_from_mixid import get_mixtures_from_mixid  # noqa: E402
from sonusai.mixture.truth_reduction import truth_reduction  # noqa: E402
from sonusai.mixture.estimate_audio_length import estimate_audio_length  # noqa: E402
from sonusai.mixture.generate_segsnr import generate_segsnr  # noqa: E402
from sonusai.mixture.generate_truth import generate_truth  # noqa: E402
from sonusai.mixture.generate_truth import truth_function  # noqa: E402
from sonusai.mixture.mixture_data import get_audio_and_truth_t  # noqa: E402
from sonusai.mixture.mixture_data import get_target_noise_audio  # noqa: E402
from sonusai.mixture.mixture_data import set_mixture_offsets  # noqa: E402
from sonusai.mixture.new_mixdb_from_mixid import new_mixdb_from_mixid  # noqa: E402
from sonusai.mixture.load_mixid import load_mixid  # noqa: E402
from sonusai.mixture.load_mixdb import load_mixdb  # noqa: E402



import numpy as np

from sonusai import SonusAIError
from sonusai import logger


def calculate_metrics(cmall: np.ndarray, binary_mode: bool = True) -> (np.ndarray, np.ndarray, np.ndarray):
    # Calculates metrics from a confusion matrix of num_classes x num_classes or
    # sets of cm S1 x S2 x ... x num_classes x num_classes where the sum is taken over
    # all the other dimensions and metrics are calculated on this sum.
    #
    # binary_mode sets binary mode where if True and num_classes=2 then only the last (-1)
    # dimension of metrics is returned (the positive metrics), i.e. a 1x10 array.
    #
    # Returns:
    # metrics  num_classes x 12            [ACC, TPR, PPV, TNR, FPR, HITFA, F1, MCC, NT, PT, TP, FP]
    # mcm      num_classes x 2 x 2
    # cmsum    num_classes x num_classes   Sum of all sum counts

    if cmall.shape[-1] != cmall.shape[-2]:
        logger.exception('Confusion matrix must be a square matrix.')
        raise SonusAIError

    num_classes = cmall.shape[-1]
    ndimsum = np.ndim(cmall) - 2
    if ndimsum == 0:
        cmsum = cmall
    else:
        dsi = tuple([i for i in range(0, cmall.ndim - 2)])
        cmsum = np.sum(cmall, dsi)  # sum over all dims except -2,-1

    metrics = np.zeros((num_classes, 12))
    mcm = np.zeros((num_classes, 2, 2))
    total_count = np.sum(np.sum(cmsum))
    eps = np.finfo(float).eps

    for nci in range(num_classes):
        mcm[nci, 1, 1] = cmsum[nci, nci]  # TP True positive
        mcm[nci, 1, 0] = np.sum(cmsum[nci,]) - mcm[nci, 1, 1]  # FN False negative = true but predicted negative
        mcm[nci, 0, 1] = np.sum(cmsum[:, nci]) - mcm[nci, 1, 1]  # FP False positive
        mcm[nci, 0, 0] = total_count - np.sum(mcm[nci,])  # TN True negative = false and predicted negative
        # True negative
        TN = mcm[nci, 0, 0]
        # False positive
        FP = mcm[nci, 0, 1]
        # False negative
        FN = mcm[nci, 1, 0]
        # True positive
        TP = mcm[nci, 1, 1]
        # Accuracy
        ACC = (TP + TN) / (TP + TN + FP + FN + eps)
        # True positive rate, sensitivity, recall, hit rate (note eps in numerator)
        # When ``true positive + false negative == 0``, recall is undefined, set to 0
        TPR = (TP) / (TP + FN + eps)
        # Precision, positive predictive value
        # When ``true positive + false positive == 0``, precision is undefined, set to 0
        PPV = TP / (TP + FP + eps)
        # Specificity i.e., selectivity, or true negative rate
        TNR = TN / (TN + FP + eps)
        # False positive rate = 1-specificity, roc x-axis
        FPR = FP / (TN + FP + eps)
        # HitFA used by some separation research, close match to MCC
        HITFA = TPR - FPR
        # F1 harmonic mean of precision, recall = 2*PPV*TPR / (PPV + TPR)
        F1 = 2 * TP / (2 * TP + FP + FN + eps)
        # Matthew correlation coefficient
        MCC = (TP * TN - FP * FN) / (np.sqrt((TP + FP) * (TP + FN) * (TN + FP) * (TN + FN)) + eps)
        # Num. negatives total (truth), also = TN+FP denom of FPR
        NT = sum(mcm[nci, 0,])
        # Num. positives total (truth), also = FN+TP denom of TPR, precision
        PT = sum(mcm[nci, 1,])
        metrics[nci] = [ACC, TPR, PPV, TNR, FPR, HITFA, F1, MCC, NT, PT, TP, FP]

    if num_classes == 2 and binary_mode:
        # If binary, drop dim0 stats which are redundant
        mcm = mcm[1:, ]
        metrics = metrics[1:, ]

    return metrics, mcm, cmsum



"""sonusai

usage: sonusai [--version] [--help] <command> [<args>...]

The sonusai commands are:
   evaluate     Evaluate model performance
   genft        Generate feature and truth data
   genmix       Generate mixture and truth data
   genmixdb     Generate a mixture database
   gentcst      Generate target configuration from a subdirectory tree
   mkwav        Make WAV files from training data
   predict      Run predict on a trained model

Aaware Sound and Voice Machine Learning Framework. See 'sonusai help <command>'
for more information on a specific command.

"""
from subprocess import call

from docopt import docopt

import sonusai
from sonusai import logger
from sonusai.utils import trim_docstring


def main():
    try:
        command_list = 'evaluate genft genmix genmixdb gentcst mkwav predict'.split()

        args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

        command = args['<command>']
        argv = args['<args>']

        if command is None:
            exit(call([__file__, '--help']))
        elif command == 'help':
            if not argv:
                exit(call([__file__, '--help']))
            elif argv[0] in command_list:
                exit(call([f'sonusai-{argv[0]}', '-h']))
            else:
                logger.error(f"{argv[0]} is not a SonusAI command. See 'sonusai help'.")
                raise SystemExit(1)
        elif command in command_list:
            exit(call([f'sonusai-{command}', *argv]))

        logger.error(f"{command} is not a SonusAI command. See 'sonusai help'.")
        raise SystemExit(1)

    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)


if __name__ == '__main__':
    main()



# SonusAI metrics utilities for model training and validation

from sonusai.metrics.averages import averages
from sonusai.metrics.calculate_class_weights import calculate_class_weights
from sonusai.metrics.calculate_metrics import calculate_metrics
from sonusai.metrics.calculate_sample_weights import calculate_sample_weights
from sonusai.metrics.confusion_matrix_to_string import confusion_matrix_to_string
from sonusai.metrics.format_confusion_matrix import format_confusion_matrix
from sonusai.metrics.generate_snr_summary import generate_snr_summary
from sonusai.metrics.generate_summary import generate_summary
from sonusai.metrics.one_hot import one_hot



import json

import h5py
import numpy as np


def read_feature_data(filename: str) -> (dict, np.ndarray, np.ndarray, np.ndarray):
    with h5py.File(name=filename, mode='r') as f:
        return json.loads(f.attrs['mixdb']), f['feature'][:], f['truth_f'][:], f['segsnr'][:]



from os.path import exists
from os.path import expandvars

import numpy as np
from tqdm import tqdm

from sonusai import SonusAIError
from sonusai import logger
from sonusai.mixture import apply_augmentation
from sonusai.mixture import read_audio


def build_noise_audio_db(mixdb: dict, show_progress: bool = False) -> list:
    db = list()
    for file_index in tqdm(sorted(list(set([sub['noise_file_index'] for sub in mixdb['mixtures']]))),
                           desc='Read noise audio', disable=not show_progress):
        audio_in = read_audio(name=mixdb['noises'][file_index]['name'], dither=mixdb['dither'])
        for augmentation_index in sorted(list(set([sub['noise_augmentation_index'] for sub in mixdb['mixtures']]))):
            db.append({
                'file':         file_index,
                'augmentation': augmentation_index,
                'audio':        apply_augmentation(audio_in=audio_in,
                                                   augmentation=mixdb['noise_augmentations'][augmentation_index],
                                                   length_common_denominator=1,
                                                   dither=mixdb['dither'])
            })
    return db


def get_noise_audio_from_db(db: list, file_index: int, augmentation_index: int) -> np.ndarray:
    for record in db:
        if record['file'] == file_index and record['augmentation'] == augmentation_index:
            return record['audio']

    logger.exception(
        f'Could not find noise_file_index {file_index} and noise_augmentation_index {augmentation_index} in noise database')
    raise SonusAIError


def build_target_audio_db(mixdb: dict, show_progress: bool = False) -> list:
    db = []
    for file_index in tqdm(sorted(list(set([sub['target_file_index'] for sub in mixdb['mixtures']]))),
                           desc='Read target audio', disable=not show_progress):
        db.append({
            'file':  file_index,
            'audio': read_audio(name=mixdb['targets'][file_index]['name'], dither=mixdb['dither'])
        })
    return db


def get_target_audio_from_db(db: list, file_index: int) -> np.ndarray:
    for record in db:
        if record['file'] == file_index:
            return record['audio']

    logger.exception(f'Could not find target_file_index {file_index} in target database')
    raise SonusAIError


def audio_files_exist(mixdb: dict) -> bool:
    result = True

    for file_index in sorted(list(set(sub['noise_file_index'] for sub in mixdb['mixtures']))):
        file_name = expandvars(mixdb['noises'][file_index]['name'])
        if not exists(file_name):
            logger.exception(f'Could not find {file_name}')
            result = False

    for file_index in sorted(list(set(sub['target_file_index'] for sub in mixdb['mixtures']))):
        file_name = expandvars(mixdb['targets'][file_index]['name'])
        if not exists(file_name):
            logger.exception(f'Could not find {file_name}')
            result = False

    return result
